/*
  hatenakeyword.h: Hatena keyword annotation program
  based on Double-Array TRIE
  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
    
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
    
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA
*/

#ifndef _HATENAKEYWORD_AUTOLINK_H_
#define _HATENAKEYWORD_AUTOLINK_H_

#include "darts.h"
#include <cstdio>
#include <cstring>

#define DEFAULT_PATTERN "<a href=\"http://d.hatena.ne.jp/keyword/%e\">%s</a>"

namespace HatenaKeyword {

  inline static char getEscapedChar (const char p) 
  {
    switch (p) {
    case '0':  return '\0'; break;
    case 'a':  return '\a'; break;
    case 'b':  return '\b'; break;
    case 't':  return '\t'; break;
    case 'n':  return '\n'; break;
    case 'v':  return '\v'; break;
    case 'f':  return '\f'; break;
    case 'r':  return '\r'; break;
    case 's':  return ' ';  break; // space          
    case '\\': return '\\'; break;
    default:  return p; break;
    }
    return '\0'; // never be here
  }
 
  class AutoLink {
  private:
    Darts::DoubleArray da;
    Darts::DoubleArray::result_pair_type daresult[512];
    AutoLink() {}; // private
    char *buf;
    char *pattern;
    size_t alloc_size;
    size_t size;
    int charset;

    void reserve(size_t);
    void write(char );
    void write(const char*);
    void write(const char*, size_t);
    size_t getmblen(const char*, const char*);
    void encodeURL(const char*, size_t);
  public:
    enum { EUC_JP, SHIFT_JIS, UTF8 };
     
    AutoLink(const char*);
    void setPattern(const char*);
    void setCharset(const char*);
    virtual ~AutoLink();
    const char *assign(const char*);
  };

  AutoLink::AutoLink(const char* filename): 
    buf(0), pattern(0), alloc_size(0), size(0), charset(EUC_JP)
  {
    if (da.open(filename) != 0) {
      std::fprintf(stderr, "no such file or directory: %s\n", filename);
      std::exit(-1);
    }
    setPattern(DEFAULT_PATTERN);
  }

  AutoLink::~AutoLink() 
  {
    delete [] buf;     buf = 0;
    delete [] pattern; pattern = 0;
    alloc_size = 0;
    size = 0;
  }

  void AutoLink::setCharset(const char* charset_)
  {
    if (std::strcmp(charset_, "EUC") == 0 ||
	std::strcmp(charset_, "EUC_JP") == 0 || 
	std::strcmp(charset_, "EUC-JP") == 0) {
      charset = EUC_JP;
      return;
    } else if (std::strcmp(charset_, "CP932") == 0 ||
	std::strcmp(charset_, "SJIS") == 0 || 
	std::strcmp(charset_, "SHIFT_JIS") == 0 ||
	std::strcmp(charset_, "SHIFT-JIS") == 0)   {
      charset = SHIFT_JIS;
      return;
    } else if (std::strcmp(charset_, "UTF8") == 0 ||
	       std::strcmp(charset_, "UTF-8") == 0) {
      charset = UTF8;
      return;
    }
    std::fprintf(stderr, "unknown charset: %s\n", charset_);
    return;
  }

  void AutoLink::setPattern(const char*pattern_) 
  {
    delete [] pattern;
    pattern = new char [strlen(pattern_) + 1];
    strcpy(pattern, pattern_);
  }
   
  size_t AutoLink::getmblen (const char*begin, const char *end) 
  {
    size_t len = end - begin;

    switch (charset) {
    case SHIFT_JIS:
      if ( (static_cast<unsigned char>(begin[0]) & 0x80) && len >= 2) {
	return 2;
      } 
      return 1;
      break;
    case UTF8:
      if (static_cast<unsigned char>(begin[0]) < 0x80) {
	return 1;
      } else if (len >= 2 && (begin[0] & 0xe0) == 0xc0) {
	return 2;
      } else if (len >= 3 && (begin[0] & 0xf0) == 0xe0) {
	return 3;
      } else if (len >= 4 && (begin[0] & 0xf8) == 0xf0) {
	return 4;
      } else if (len >= 5 && (begin[0] & 0xfc) == 0xf8) {
	return 5;
      } else if (len >= 6 && (begin[0] & 0xfe) == 0xfc) {
	return 6;
      } else {
	return 1;
      }
      break;
    case EUC_JP:
      if (static_cast<unsigned char>(begin[0]) == 0x8f && len >= 3) { // JISX 0212, 0213
	unsigned short key = (static_cast<unsigned char>(begin[1]) << 8) 
	  + static_cast<unsigned char>(begin[2]);
	if (key < 0xA0A0) { // offset  violation
	  return 1;
	}
	return 3;
	// JISX 0208 + 0201
      } else if ( (static_cast<unsigned char>(begin[0]) & 0x80) && len >= 2) { 
	return 2;
      } else {
	return 1;
      }
      break;
    }

    return 1;
  }

  void AutoLink::reserve (size_t _l)
  {
    if (size + _l >= alloc_size) {
      if (alloc_size == 0) {
	alloc_size = 8192;
	buf = new char [alloc_size];
      }
      size_t len = size + _l;
      do { alloc_size *= 2; } while (len >= alloc_size);
      char *new_buf = new char [alloc_size];
      memcpy (new_buf, buf, size);
      delete [] buf;
      buf = new_buf;
    }
  }

  void AutoLink::write (char _n)
  {
    reserve (1);
    buf[size] = _n;
    ++size;
  }

  void AutoLink::write(const char *_n)
  {
    this->write (_n, strlen(_n));
  }

  void AutoLink::write(const char* _n, size_t _l)
  {
    reserve (_l);
    memcpy (buf + size , _n, _l);
    size += _l;
  }

  void AutoLink::encodeURL(const char *begin, size_t size)
  {
    static const char  *digits = "0123456789abcdef";
    const char *end = begin + size;
    for (const char *p = begin; p != end; ++p) {
      if (isascii(*p) && (isdigit(*p) || isalpha(*p)))
	this->write(*p);
      else {
	this->write('%');
	this->write(digits[(*p >> 4) & 0x0f]);
	this->write(digits[*p & 0x0f]);
      }
    }
  }

  const char *AutoLink::assign(const char* text) 
  {
    const char *begin = text;
    const char *end   = text + strlen(text);

    size = 0;

    while (begin < end) {

      size_t size = da.commonPrefixSearch(begin, daresult, 
					  sizeof(daresult),
					  (size_t)(end - begin));
      size_t seekto = 0;       

      if (size) {
	 
	for (size_t i = 0; i < size; ++i) {
	  if (seekto < daresult[i].length)
	    seekto = daresult[i].length;
	}
	 
	if (seekto) {

	  for (char *p = pattern; *p; p++) {
	    switch (*p) {
	    default:   this->write(*p); break;
	    case '\\': this->write(getEscapedChar(*++p)); break;
	    case '%': 
	      switch (*++p) {
	      case 's': this->write(begin, seekto); break;
	      case 'e': this->encodeURL(begin, seekto); break;
	      default: break; // do nothing
	      }
	    }
	  }

	  begin += seekto;
	}
      }
       
      if (seekto == 0) {
	size_t len = getmblen(begin, end);
	this->write(begin, len);
	begin += len;
      }
    }

    this->write('\0');

    return const_cast<const char*>(buf);
  }
}
#endif
  
